from typing import List

import numpy as np
from OpenGL.GL import GL_QUADS, glGenVertexArrays, glBindVertexArray, \
    glGenBuffers, glBindBuffer, GL_ARRAY_BUFFER, glBufferData, \
    GL_STATIC_DRAW, glVertexAttribPointer, GL_FLOAT, GL_FALSE, GL_ELEMENT_ARRAY_BUFFER, \
    glEnableVertexAttribArray, glDrawElements, GL_UNSIGNED_INT, glDisableVertexAttribArray, glGenTextures, GL_UNPACK_ALIGNMENT, glPixelStorei, \
    GL_TEXTURE_2D, \
    glBindTexture, glTexParameteri, GL_TEXTURE_BASE_LEVEL, GL_TEXTURE_MAX_LEVEL, glTexImage2D, GL_UNSIGNED_BYTE, \
    GL_RGBA, GL_TEXTURE_MIN_FILTER, GL_TEXTURE_MAG_FILTER, GL_NEAREST, glDeleteBuffers, \
    glDeleteVertexArrays, glDeleteTextures
from PIL import Image

from gui.Layer import Layer
from gui.opengl import OpenGLGUIFacade


class OpenGLLayer(Layer):

    def __init__(self, gui: OpenGLGUIFacade):
        self.__gui = gui

        # Texture properties
        self.__textureId = -1  # type: int
        self.__textureWidth = 0  # type: int
        self.__textureHeight = 0  # type: int

        # Tile properties
        self.__tileWidth = 0  # type: int
        self.__tileHeight = 0  # type: int
        self.__textureTileWidth = 0  # type: float
        self.__textureTileHeight = 0  # type: float
        self.__screenTileWidth = 0  # type: float
        self.__screenTileHeight = 0  # type: float

        # Mesh properties
        self.__vaoId = -1  # type: int
        self.__vboIds = []  # type: List[int]
        self.__faceCount = 0  # type: int

    @property
    def tileWidth(self) -> int:
        return self.__tileWidth

    @property
    def tileHeight(self) -> int:
        return self.__tileHeight

    @property
    def textureWidth(self) -> int:
        return self.__textureWidth

    @property
    def textureHeight(self) -> int:
        return self.__textureHeight

    @property
    def screenTileWidth(self) -> float:
        return self.__screenTileWidth

    @property
    def screenTileHeight(self) -> float:
        return self.__screenTileHeight

    def setTexture(self, fileName: str, tileWidth: int, tileHeight: int):
        # Load texture image
        image = Image.open(fileName)
        assert image.mode == "RGBA"
        imageArray = np.array(image)

        # Create texture from image
        self.__textureId = glGenTextures(1)
        glPixelStorei(GL_UNPACK_ALIGNMENT, 4)
        glBindTexture(GL_TEXTURE_2D, self.__textureId)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_BASE_LEVEL, 0)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_LEVEL, 0)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, image.size[0], image.size[1],
                     0, GL_RGBA, GL_UNSIGNED_BYTE, imageArray)

        # Tile properties
        self.__tileWidth = tileWidth
        self.__tileHeight = tileHeight
        self.__textureWidth = image.size[0]
        self.__textureHeight = image.size[1]
        self.__textureTileWidth = float(tileWidth) / float(self.__textureWidth)
        self.__textureTileHeight = float(tileHeight) / float(self.__textureHeight)
        self.__screenTileWidth = 2 * float(tileWidth) / float(self.__gui.screenWidth)
        self.__screenTileHeight = 2 * float(tileHeight) / float(self.__gui.screenHeight)

    def setGridMesh(self, grid: np.ndarray):
        assert grid.ndim == 3
        gridWidth = grid.shape[0]
        gridHeight = grid.shape[1]
        assert grid.shape[2] == 2

        vertices = np.empty([gridWidth, gridHeight, 4, 2], dtype=np.float32)
        uvMap = np.empty([gridWidth, gridHeight, 4, 2], dtype=np.float32)
        faces = np.empty([gridWidth, gridHeight, 4], dtype=np.uint)
        faceCount = 0
        for y in range(gridHeight):
            for x in range(gridWidth):
                spriteScreenX1 = -1 + x * self.__screenTileWidth
                spriteScreenY1 = 1 - y * self.__screenTileHeight
                spriteScreenX2 = spriteScreenX1 + self.__screenTileWidth
                spriteScreenY2 = spriteScreenY1 - self.__screenTileHeight
                vertices[x, y, 0] = [spriteScreenX1, spriteScreenY2]
                vertices[x, y, 1] = [spriteScreenX1, spriteScreenY1]
                vertices[x, y, 2] = [spriteScreenX2, spriteScreenY1]
                vertices[x, y, 3] = [spriteScreenX2, spriteScreenY2]

                spriteTextureX1 = grid[x, y, 0] * self.__textureTileWidth
                spriteTextureY1 = grid[x, y, 1] * self.__textureTileHeight
                spriteTextureX2 = spriteTextureX1 + self.__textureTileWidth
                spriteTextureY2 = spriteTextureY1 + self.__textureTileHeight
                uvMap[x, y, 0] = [spriteTextureX1, spriteTextureY2]
                uvMap[x, y, 1] = [spriteTextureX1, spriteTextureY1]
                uvMap[x, y, 2] = [spriteTextureX2, spriteTextureY1]
                uvMap[x, y, 3] = [spriteTextureX2, spriteTextureY2]

                faces[x, y, 0] = faceCount * 4
                faces[x, y, 1] = faceCount * 4 + 1
                faces[x, y, 2] = faceCount * 4 + 2
                faces[x, y, 3] = faceCount * 4 + 3
                faceCount += 1

        self.setMeshData(vertices, faces, uvMap)

    def setMeshData(self, vertices: np.ndarray, faces: np.ndarray, uvMap: np.ndarray):
        assert vertices.ndim == 4
        gridWidth = vertices.shape[0]
        gridHeight = vertices.shape[1]
        assert vertices.shape[2] == 4
        assert vertices.shape[3] == 2
        assert vertices.dtype == np.float32

        assert faces.ndim == 3
        assert faces.shape[0] == gridWidth
        assert faces.shape[1] == gridHeight
        assert faces.shape[2] == 4
        assert faces.dtype == np.uint

        assert uvMap.ndim == 4
        assert uvMap.shape[0] == gridWidth
        assert uvMap.shape[1] == gridHeight
        assert uvMap.shape[2] == 4
        assert uvMap.shape[3] == 2
        assert uvMap.dtype == np.float32

        # Create one Vertex Array Object (VAO)
        self.__vaoId = glGenVertexArrays(1)
        # We will be working on this VAO
        glBindVertexArray(self.__vaoId)

        # Create three Vertex Buffer Objects (VBO)
        self.__vboIds = glGenBuffers(3)
        # Commodity variables to memorize the id of each VBO
        vertexVboId = self.__vboIds[0]
        uvVboId = self.__vboIds[1]
        indexVboId = self.__vboIds[2]

        # Copy vertices data to GPU
        glBindBuffer(GL_ARRAY_BUFFER, vertexVboId)
        vertices = np.ascontiguousarray(vertices.flatten())
        glBufferData(GL_ARRAY_BUFFER, 4 * len(vertices), vertices, GL_STATIC_DRAW)
        glVertexAttribPointer(0, 2, GL_FLOAT, GL_FALSE, 0, None)

        # Copy UV coords data to GPU
        glBindBuffer(GL_ARRAY_BUFFER, uvVboId)
        uvMap = np.ascontiguousarray(uvMap.flatten())
        glBufferData(GL_ARRAY_BUFFER, 4 * len(uvMap), uvMap, GL_STATIC_DRAW)
        glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, 0, None)

        # Copy index data to GPU
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, indexVboId)
        faces = np.ascontiguousarray(faces.flatten())
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, 4 * len(faces), faces, GL_STATIC_DRAW)
        self.__faceCount = gridWidth * gridHeight

    def draw(self):
        glBindVertexArray(self.__vaoId)
        glBindTexture(GL_TEXTURE_2D, self.__textureId)
        glEnableVertexAttribArray(0)
        glEnableVertexAttribArray(1)
        glDrawElements(GL_QUADS, 4 * self.__faceCount, GL_UNSIGNED_INT, None)
        glDisableVertexAttribArray(0)
        glDisableVertexAttribArray(1)
        glBindVertexArray(0)

    def deleteObjects(self):
        glDeleteBuffers(3, self.__vboIds)
        glDeleteTextures(1, [self.__textureId])
        glDeleteVertexArrays(1, [self.__vaoId])
